"""This module handles all settings related to sound output."""
from __future__ import annotations

import re
import subprocess
import time
from typing import Dict, TYPE_CHECKING, Optional, List

from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import JsonResponse

from core import util
from core.models import Setting
from core.settings.settings import Settings
from core.util import background_thread


class Sound:
    """This class is responsible for handling settings changes related to sound output."""

    def __init__(self, settings: Settings):
        self.settings = settings
        self.output = Settings.get_setting("sound_output", "")
        self.backup_stream = Settings.get_setting("backup_stream", "")
        self.bluetoothctl: Optional[subprocess.Popen[bytes]] = None
        self.bluetooth_devices: List[Dict[str, str]] = []

    @Settings.option
    def set_voting_system(self, request: WSGIRequest) -> None:
        """Enables or disables the voting system based on the given value."""
        enabled = request.POST.get("value") == "true"
        Setting.objects.filter(key="voting_system").update(value=enabled)
        self.voting_system = enabled

    @Settings.option
    def set_backup_stream(self, request: WSGIRequest) -> None:
        """Sets the given internet stream as backup stream."""
        stream = request.POST.get("value")
        Setting.objects.filter(key="backup_stream").update(value=stream)
        self.backup_stream = stream

    def _get_bluetoothctl_line(self) -> str:
        # Note: this variable is not guarded by a lock.
        # But there should only be one admin accessing these bluetooth functions anyway.
        if self.bluetoothctl is None:
            return ""
        assert self.bluetoothctl.stdout
        line = self.bluetoothctl.stdout.readline().decode()
        ansi_escape = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")
        line = ansi_escape.sub("", line)
        line = line.strip()
        return line

    def _stop_bluetoothctl(self) -> None:
        if self.bluetoothctl:
            assert self.bluetoothctl.stdin
            self.bluetoothctl.stdin.close()
            self.bluetoothctl.wait()
        self.bluetoothctl = None

    @Settings.option
    def set_bluetooth_scanning(self, request: WSGIRequest) -> HttpResponse:
        """Enables scanning of bluetooth devices."""
        enabled = request.POST.get("value") == "true"
        if enabled:
            if self.bluetoothctl is not None:
                return HttpResponseBadRequest("Already Scanning")
            self.bluetooth_devices = []
            self.bluetoothctl = subprocess.Popen(
                ["bluetoothctl"], stdin=subprocess.PIPE, stdout=subprocess.PIPE
            )
            assert self.bluetoothctl.stdin

            self.bluetoothctl.stdin.write(b"devices\n")
            self.bluetoothctl.stdin.write(b"scan on\n")
            self.bluetoothctl.stdin.flush()

            @background_thread
            def do_scan():
                while True:
                    line = self._get_bluetoothctl_line()
                    if not line:
                        break
                    # match old devices
                    match = re.match(r"Device (\S*) (.*)", line)
                    # match newly scanned devices
                    # We need the '.*' at the beginning of the line to account for control sequences
                    if not match:
                        match = re.match(r".*\[NEW\] Device (\S*) (.*)", line)
                    if match:
                        address = match.group(1)
                        name = match.group(2)
                        # filter unnamed devices
                        # devices named after their address are no speakers
                        if re.match("[A-Z0-9][A-Z0-9](-[A-Z0-9][A-Z0-9]){5}", name):
                            continue
                        self.bluetooth_devices.append(
                            {"address": address, "name": name}
                        )
                        self.settings.update_state()

            do_scan()
            return HttpResponse("Started scanning")
        else:
            if self.bluetoothctl is None:
                return HttpResponseBadRequest("Currently not scanning")
            self._stop_bluetoothctl()
            return HttpResponse("Stopped scanning")

    @Settings.option
    def connect_bluetooth(self, request: WSGIRequest) -> HttpResponse:
        """Connect to a given bluetooth device."""
        address = request.POST.get("address")
        if self.bluetoothctl is not None:
            return HttpResponseBadRequest("Stop scanning before connecting")
        if address is None or address == "":
            return HttpResponseBadRequest("No device selected")

        self.bluetoothctl = subprocess.Popen(
            ["bluetoothctl"], stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        assert self.bluetoothctl.stdin
        error = ""

        # A Function that acts as a timeout for unexpected errors (or timeouts)
        @background_thread
        def _timeout() -> None:
            nonlocal error
            time.sleep(20)
            error = "Timed out"
            if self.bluetoothctl is not None:
                self._stop_bluetoothctl()

        # Sometimes, pairing hangs forever. Since connecting alone is enough, skip pairing.
        # self.bluetoothctl.stdin.write(b"pair " + address.encode() + b"\n")
        # self.bluetoothctl.stdin.flush()
        # while True:
        #     line = self._get_bluetoothctl_line()
        #     if not line:
        #         break
        #     if re.match(".*Device " + address + " not available", line):
        #         error = "Device unavailable"
        #         break
        #     if re.match(".*Failed to pair: org.bluez.Error.AlreadyExists", line):
        #         break
        #     if re.match(".*Pairing successful", line):
        #         break

        # if error:
        #     self._stop_bluetoothctl()
        #     return HttpResponseBadRequest(error)

        self.bluetoothctl.stdin.write(b"connect " + address.encode() + b"\n")
        self.bluetoothctl.stdin.flush()
        while True:
            line = self._get_bluetoothctl_line()
            if not line:
                break
            if re.match(".*Device " + address + " not available", line):
                error = "Device unavailable"
                break
            if re.match(".*Failed to connect: org.bluez.Error.Failed", line):
                error = "Connect Failed"
                break
            if re.match(".*Failed to connect: org.bluez.Error.InProgress", line):
                error = "Connect in progress"
                break
            if re.match(".*Connection successful", line):
                break
        # trust the device to automatically reconnect when it is available again
        self.bluetoothctl.stdin.write(b"trust " + address.encode() + b"\n")
        self.bluetoothctl.stdin.flush()

        self._stop_bluetoothctl()
        if error:
            return HttpResponseBadRequest(error)

        return HttpResponse("Connected. Set output device to activate.")

    @Settings.option
    def disconnect_bluetooth(self, request: WSGIRequest) -> HttpResponse:
        """Untrusts a given bluetooth device to prevent automatic reconnects.
        Does not unpair or remove the device."""
        address = request.POST.get("address")
        if self.bluetoothctl is not None:
            return HttpResponseBadRequest("Stop scanning before disconnecting")
        if address is None or address == "":
            return HttpResponseBadRequest("No device selected")

        self.bluetoothctl = subprocess.Popen(
            ["bluetoothctl"], stdin=subprocess.PIPE, stdout=subprocess.PIPE
        )
        assert self.bluetoothctl.stdin
        error = ""

        self.bluetoothctl.stdin.write(b"untrust " + address.encode() + b"\n")
        self.bluetoothctl.stdin.flush()
        while True:
            line = self._get_bluetoothctl_line()
            if not line:
                break
            if re.match(".*Device " + address + " not available", line):
                error = "Device unavailable"
                break
            if re.match(".*untrust succeeded", line):
                break

        self._stop_bluetoothctl()
        if error:
            return HttpResponseBadRequest(error)
        return HttpResponse("Disconnected")

    @Settings.option
    def list_outputs(self, _request: WSGIRequest) -> JsonResponse:
        """Returns a list of all sound output devices currently available."""
        output = subprocess.check_output(
            "pactl list short sinks".split(),
            env={"PULSE_SERVER": "127.0.0.1"},
            universal_newlines=True,
        )
        tokenized_lines = [line.split() for line in output.splitlines()]

        sinks = ["fakesink", "icecast", "snapcast"]
        sinks.extend([sink[1] for sink in tokenized_lines if len(sink) >= 2])

        return JsonResponse(sinks, safe=False)

    def _set_output(self, output: str) -> HttpResponse:
        icecast_installed = util.service_installed("icecast2")
        snapcast_installed = util.service_installed("snapserver")

        if output == "fakesink":
            mopidy_output = "fakesink"
        elif output == "icecast":
            if not icecast_installed:
                return HttpResponseBadRequest("Please install icecast2")

            subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_icecast"])
            mopidy_output = "icecast"
        elif output == "snapcast":
            if not snapcast_installed:
                return HttpResponseBadRequest("Please install snapserver")

            subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_snapcast"])
            mopidy_output = "snapcast"
        else:
            try:
                subprocess.run(
                    ["pactl", "set-default-sink", output],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.PIPE,
                    env={"PULSE_SERVER": "127.0.0.1"},
                    check=True,
                )
                mopidy_output = "pulse"
            except subprocess.CalledProcessError as e:
                return HttpResponseBadRequest(e.stderr)

        if icecast_installed and output != "icecast":
            subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_icecast"])
        if snapcast_installed and output != "snapcast":
            subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_snapcast"])

        self.settings.system.update_mopidy_config(mopidy_output)

        return HttpResponse(
            "Output was set. Restarting the current song might be necessary."
        )

    @Settings.option
    def set_output(self, request: WSGIRequest) -> HttpResponse:
        """Sets the given device as default output device."""
        output = request.POST.get("value")
        if not output:
            return HttpResponseBadRequest("No output selected")

        if output == self.output:
            return HttpResponseBadRequest("Output unchanged")

        Setting.objects.filter(key="sound_output").update(value=output)
        self.output = output

        return self._set_output(output)



"""This module contains everything related to the settings and configuration of the server."""
# pylint: disable=no-self-use  # self is used in decorator

from __future__ import annotations

import os
import subprocess
import yaml
from functools import wraps
from typing import Callable, Dict, Any, TYPE_CHECKING, Optional, TypeVar, List

from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest

# JsonResponse needs to be imported here so types can resolved through the decorator?
from django.http import HttpResponse, JsonResponse
from django.http import HttpResponseForbidden
from django.shortcuts import render, redirect
from django.urls import URLPattern

from core.models import Setting
from core.state_handler import Stateful

if TYPE_CHECKING:
    from core.base import Base
    from core.settings.basic import Basic
    from core.settings.platforms import Platforms
    from core.settings.sound import Sound
    from core.settings.wifi import Wifi
    from core.settings.library import Library
    from core.settings.analysis import Analysis
    from core.settings.system import System

    T = TypeVar(  # pylint: disable=invalid-name
        "T", Basic, Platforms, Sound, Wifi, Library, Analysis, System
    )


class Settings(Stateful):
    """This class is responsible for handling requests from the /settings page."""

    @staticmethod
    def get_setting(key: str, default: str) -> str:
        """This method returns the value for the given :param key:.
        Vaules of non-existing keys are set to :param default:"""
        if settings.MOCK:
            return default
        else:
            return Setting.objects.get_or_create(key=key, defaults={"value": default})[
                0
            ].value

    @staticmethod
    def option(
        func: Callable[[T, WSGIRequest], Optional[HttpResponse]]
    ) -> Callable[[T, WSGIRequest], HttpResponse]:
        """A decorator that makes sure that only the admin changes a setting."""

        def _decorator(self: T, request: WSGIRequest) -> HttpResponse:
            if not self.settings.base.user_manager.is_admin(request.user):
                return HttpResponseForbidden()
            response = func(self, request)
            self.settings.update_state()
            if response is not None:
                return response
            return HttpResponse()

        return wraps(func)(_decorator)

    def __init__(self, base: "Base") -> None:
        from core.settings.basic import Basic
        from core.settings.platforms import Platforms
        from core.settings.sound import Sound
        from core.settings.wifi import Wifi
        from core.settings.library import Library
        from core.settings.analysis import Analysis
        from core.settings.system import System

        self.base = base
        self.urlpatterns: List[URLPattern] = []
        self.basic = Basic(self)
        self.platforms = Platforms(self)
        self.sound = Sound(self)
        self.wifi = Wifi(self)
        self.library = Library(self)
        self.analysis = Analysis(self)
        self.system = System(self)

    def state_dict(self) -> Dict[str, Any]:
        state_dict = self.base.state_dict()

        settings_state = {}
        settings_state["votingSystem"] = self.basic.voting_system
        settings_state["newMusicOnly"] = self.basic.new_music_only
        settings_state["loggingEnabled"] = self.basic.logging_enabled
        settings_state["embedStream"] = self.basic.embed_stream
        settings_state["hashtagsActive"] = self.basic.hashtags_active
        settings_state["onlineSuggestions"] = self.basic.online_suggestions
        settings_state["numberOfSuggestions"] = self.basic.number_of_suggestions
        settings_state["peopleToParty"] = self.basic.people_to_party
        settings_state["alarmProbability"] = self.basic.alarm_probability
        settings_state["downvotesToKick"] = self.basic.downvotes_to_kick
        settings_state["maxDownloadSize"] = self.basic.max_download_size
        settings_state["additionalKeywords"] = self.basic.additional_keywords
        settings_state["forbiddenKeywords"] = self.basic.forbidden_keywords
        settings_state["maxPlaylistItems"] = self.basic.max_playlist_items
        settings_state["maxQueueLength"] = self.basic.max_queue_length
        settings_state["hasInternet"] = self.basic.has_internet

        settings_state["youtubeEnabled"] = self.platforms.youtube_enabled
        settings_state["youtubeSuggestions"] = self.platforms.youtube_suggestions

        settings_state["spotifyEnabled"] = self.platforms.spotify_enabled
        settings_state["spotifySuggestions"] = self.platforms.spotify_suggestions

        settings_state["soundcloudEnabled"] = self.platforms.soundcloud_enabled
        settings_state["soundcloudSuggestions"] = self.platforms.soundcloud_suggestions

        settings_state["jamendoEnabled"] = self.platforms.jamendo_enabled
        settings_state["jamendoSuggestions"] = self.platforms.jamendo_suggestions

        settings_state["backupStream"] = self.sound.backup_stream

        settings_state["bluetoothScanning"] = self.sound.bluetoothctl is not None
        settings_state["bluetoothDevices"] = self.sound.bluetooth_devices

        settings_state["output"] = self.sound.output

        try:
            with open(os.path.join(settings.BASE_DIR, "config/homewifi")) as f:
                settings_state["homewifiSsid"] = f.read()
        except FileNotFoundError:
            settings_state["homewifiSsid"] = ""

        settings_state["scanProgress"] = self.library.scan_progress

        try:
            settings_state["homewifiEnabled"] = (
                subprocess.call(["/usr/local/sbin/raveberry/homewifi_enabled"]) != 0
            )
            settings_state["eventsEnabled"] = (
                subprocess.call(["/usr/local/sbin/raveberry/events_enabled"]) != 0
            )
            settings_state["hotspotEnabled"] = (
                subprocess.call(["/usr/local/sbin/raveberry/hotspot_enabled"]) != 0
            )
            settings_state["wifiProtectionEnabled"] = (
                subprocess.call(["/usr/local/sbin/raveberry/wifi_protection_enabled"])
                != 0
            )
            settings_state["tunnelingEnabled"] = (
                subprocess.call(["sudo", "/usr/local/sbin/raveberry/tunneling_enabled"])
                != 0
            )
            settings_state["remoteEnabled"] = (
                subprocess.call(["/usr/local/sbin/raveberry/remote_enabled"]) != 0
            )
        except FileNotFoundError:
            settings_state["systemInstall"] = False
        else:
            settings_state["systemInstall"] = True
            with open(os.path.join(settings.BASE_DIR, "config/raveberry.yaml")) as f:
                config = yaml.safe_load(f)
            settings_state["hotspotConfigured"] = config["hotspot"]
            settings_state["remoteConfigured"] = config["remote_key"] is not None

        settings_state["youtubeConfigured"] = self.platforms.youtube_available
        settings_state["spotifyConfigured"] = self.platforms.spotify_available
        settings_state["soundcloudConfigured"] = self.platforms.soundcloud_available
        settings_state["jamendoConfigured"] = self.platforms.jamendo_available

        state_dict["settings"] = settings_state
        return state_dict

    def index(self, request: WSGIRequest) -> HttpResponse:
        """Renders the /settings page. Only admin is allowed to see this page."""
        if not self.base.user_manager.is_admin(request.user):
            return redirect("login")
        context = self.base.context(request)
        context["urls"] = self.urlpatterns
        library_path = self.library.get_library_path()
        if os.path.islink(library_path):
            context["local_library"] = os.readlink(library_path)
        else:
            context["local_library"] = "/"
        context["version"] = settings.VERSION
        return render(request, "settings.html", context)



"""This module handles all settings regarding the music platforms."""
from __future__ import annotations

import importlib
import os
import subprocess
from typing import TYPE_CHECKING

from django.conf import settings as conf
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.http import HttpResponseBadRequest

from core.models import Setting
from core.settings.library import Library
from core.settings.settings import Settings


class Platforms:
    """This class is responsible for handling setting changes related to music platforms."""

    def __init__(self, settings: Settings):
        self.settings = settings

        self.local_enabled = os.path.islink(Library.get_library_path())

        # in the docker container all dependencies are installed
        self.youtube_available = (
            conf.DOCKER or importlib.util.find_spec("youtube_dl") is not None
        )
        # the _enabled check is the second expression so the databes entry gets created in any case
        self.youtube_enabled = (
            Settings.get_setting("youtube_enabled", "True") == "True"
            and self.youtube_available
        )
        self.youtube_suggestions = int(Settings.get_setting("youtube_suggestions", "2"))

        # Spotify has no python dependencies we could easily check.
        try:
            self.spotify_available = (
                conf.DOCKER
                or "[spotify]"
                in subprocess.check_output(
                    ["mopidy", "config"], stderr=subprocess.DEVNULL
                )
                .decode()
                .splitlines()
            )
        except FileNotFoundError:
            # mopidy is not installed. Disable when mocking, enable otherwise
            self.spotify_available = not conf.MOCK
        self.spotify_enabled = (
            Settings.get_setting("spotify_enabled", "False") == "True"
            and self.spotify_available
        )
        self.spotify_suggestions = int(Settings.get_setting("spotify_suggestions", "2"))
        self.spotify_username = Settings.get_setting("spotify_username", "")
        self.spotify_password = Settings.get_setting("spotify_password", "")
        self.spotify_client_id = Settings.get_setting("spotify_client_id", "")
        self.spotify_client_secret = Settings.get_setting("spotify_client_secret", "")

        self.soundcloud_available = (
            conf.DOCKER or importlib.util.find_spec("soundcloud") is not None
        )
        self.soundcloud_enabled = (
            Settings.get_setting("soundcloud_enabled", "False") == "True"
            and self.soundcloud_available
        )
        self.soundcloud_suggestions = int(
            Settings.get_setting("soundcloud_suggestions", "2")
        )
        self.soundcloud_auth_token = Settings.get_setting("soundcloud_auth_token", "")

        # Jamendo has no python dependencies we could easily check.
        try:
            self.jamendo_available = (
                conf.DOCKER
                or "[jamendo]"
                in subprocess.check_output(
                    ["mopidy", "config"], stderr=subprocess.DEVNULL
                )
                .decode()
                .splitlines()
            )
        except FileNotFoundError:
            # mopidy is not installed. Disable when mocking, enable otherwise
            self.jamendo_available = not conf.MOCK
        self.jamendo_enabled = (
            Settings.get_setting("jamendo_enabled", "False") == "True"
            and self.jamendo_available
        )
        self.jamendo_suggestions = int(Settings.get_setting("jamendo_suggestions", "2"))
        self.jamendo_client_id = Settings.get_setting("jamendo_client_id", "")

    @Settings.option
    def set_youtube_enabled(self, request: WSGIRequest):
        """Enables or disables youtube to be used as a song provider."""
        enabled = request.POST.get("value") == "true"
        Setting.objects.filter(key="youtube_enabled").update(value=enabled)
        self.youtube_enabled = enabled

    @Settings.option
    def set_youtube_suggestions(self, request: WSGIRequest):
        """Sets the number of online suggestions from youtube to be shown."""
        value = int(request.POST.get("value"))  # type: ignore
        Setting.objects.filter(key="youtube_suggestions").update(value=value)
        self.youtube_suggestions = value

    def _set_extension_enabled(self, extension, enabled) -> HttpResponse:
        if enabled:
            if conf.DOCKER:
                response = HttpResponse(
                    "Make sure you provided mopidy with correct credentials."
                )
            else:
                extensions = self.settings.system.check_mopidy_extensions()
                functional, message = extensions[extension]
                if not functional:
                    return HttpResponseBadRequest(message)
                response = HttpResponse(message)
        else:
            response = HttpResponse("Disabled extension")
        Setting.objects.filter(key=f"{extension}_enabled").update(value=enabled)
        setattr(self, f"{extension}_enabled", enabled)
        return response

    @Settings.option
    def set_spotify_enabled(self, request: WSGIRequest) -> HttpResponse:
        """Enables or disables spotify to be used as a song provider.
        Makes sure mopidy has correct spotify configuration."""
        enabled = request.POST.get("value") == "true"
        return self._set_extension_enabled("spotify", enabled)

    @Settings.option
    def set_spotify_suggestions(self, request: WSGIRequest):
        """Sets the number of online suggestions from spotify to be shown."""
        value = int(request.POST.get("value"))  # type: ignore
        Setting.objects.filter(key="spotify_suggestions").update(value=value)
        self.spotify_suggestions = value

    @Settings.option
    def set_spotify_credentials(self, request: WSGIRequest) -> HttpResponse:
        """Update spotify credentials."""
        username = request.POST.get("username")
        password = request.POST.get("password")
        client_id = request.POST.get("client_id")
        client_secret = request.POST.get("client_secret")

        if not username or not password or not client_id or not client_secret:
            return HttpResponseBadRequest("All fields are required")

        self.spotify_username = username
        self.spotify_password = password
        self.spotify_client_id = client_id
        self.spotify_client_secret = client_secret

        Setting.objects.filter(key="spotify_username").update(
            value=self.spotify_username
        )
        Setting.objects.filter(key="spotify_password").update(
            value=self.spotify_password
        )
        Setting.objects.filter(key="spotify_client_id").update(
            value=self.spotify_client_id
        )
        Setting.objects.filter(key="spotify_client_secret").update(
            value=self.spotify_client_secret
        )

        self.settings.system.update_mopidy_config("pulse")
        return HttpResponse("Updated credentials")

    @Settings.option
    def set_soundcloud_enabled(self, request: WSGIRequest) -> HttpResponse:
        """Enables or disables soundcloud to be used as a song provider.
        Makes sure mopidy has correct soundcloud configuration."""
        enabled = request.POST.get("value") == "true"
        return self._set_extension_enabled("soundcloud", enabled)

    @Settings.option
    def set_soundcloud_suggestions(self, request: WSGIRequest):
        """Sets the number of online suggestions from soundcloud to be shown."""
        value = int(request.POST.get("value"))  # type: ignore
        Setting.objects.filter(key="soundcloud_suggestions").update(value=value)
        self.soundcloud_suggestions = value

    @Settings.option
    def set_soundcloud_credentials(self, request: WSGIRequest) -> HttpResponse:
        """Update soundcloud credentials."""
        auth_token = request.POST.get("auth_token")

        if not auth_token:
            return HttpResponseBadRequest("All fields are required")

        self.soundcloud_auth_token = auth_token

        Setting.objects.filter(key="soundcloud_auth_token").update(
            value=self.soundcloud_auth_token
        )

        self.settings.system.update_mopidy_config("pulse")
        return HttpResponse("Updated credentials")

    @Settings.option
    def set_jamendo_enabled(self, request: WSGIRequest) -> HttpResponse:
        """Enables or disables jamendo to be used as a song provider.
        Makes sure mopidy has correct jamendo configuration."""
        enabled = request.POST.get("value") == "true"
        return self._set_extension_enabled("jamendo", enabled)

    @Settings.option
    def set_jamendo_suggestions(self, request: WSGIRequest):
        """Sets the number of online suggestions from jamendo to be shown."""
        value = int(request.POST.get("value"))  # type: ignore
        Setting.objects.filter(key="jamendo_suggestions").update(value=value)
        self.jamendo_suggestions = value

    @Settings.option
    def set_jamendo_credentials(self, request: WSGIRequest) -> HttpResponse:
        """Update jamendo credentials."""
        client_id = request.POST.get("client_id")

        if not client_id:
            return HttpResponseBadRequest("All fields are required")

        self.jamendo_client_id = client_id

        Setting.objects.filter(key="jamendo_client_id").update(
            value=self.jamendo_client_id
        )

        self.settings.system.update_mopidy_config("pulse")
        return HttpResponse("Updated credentials")



from typing import TYPE_CHECKING

import ipware
import re
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, HttpResponseBadRequest
from django.views.decorators.csrf import csrf_exempt

if TYPE_CHECKING:
    from core.base import Base


class Api:
    def __init__(self, base: "Base") -> None:
        self.base = base

    @csrf_exempt
    def post_song(self, request: WSGIRequest) -> HttpResponse:
        """This endpoint is part of the API and exempt from CSRF checks.
        Shareberry uses this endpoint."""
        # only get ip on user requests
        if self.base.settings.basic.logging_enabled:
            request_ip, _ = ipware.get_client_ip(request)
            if request_ip is None:
                request_ip = ""
        else:
            request_ip = ""
        query = request.POST.get("query")
        if not query:
            return HttpResponseBadRequest("No query to share.")

        match = re.search("(?P<url>https?://[^\s]+)", query)
        if match:
            query = match.group("url")

        # Set the requested platform to 'spotify'.
        # It will automatically fall back to Youtube
        # if Spotify is not enabled or a youtube link was requested.
        successful, message, _ = self.base.musiq.do_request_music(
            request_ip, query, None, False, "spotify"
        )
        if not successful:
            return HttpResponseBadRequest(message)
        return HttpResponse(message)



"""This module handles all settings related to system configuration."""
from __future__ import annotations

import configparser
import os
import re
import shutil
import subprocess
import time
from typing import Dict, Tuple, Optional

import cachetools.func
import requests
from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse, JsonResponse
from django.http import HttpResponseBadRequest

from core.settings.settings import Settings


class System:
    """This class is responsible for handling settings changes related to system configuration."""

    @staticmethod
    def update_mopidy_config(output: str) -> None:
        """Updates mopidy's config with the credentials stored in the database.
        If no config_file is given, the default one is used."""
        if settings.DOCKER:
            # raveberry cannot restart mopidy in the docker setup
            return

        if output == "pulse":
            if shutil.which("cava"):
                output = "cava"
            else:
                output = "regular"

        spotify_username = Settings.get_setting("spotify_username", "")
        spotify_password = Settings.get_setting("spotify_password", "")
        spotify_client_id = Settings.get_setting("spotify_client_id", "")
        spotify_client_secret = Settings.get_setting("spotify_client_secret", "")
        soundcloud_auth_token = Settings.get_setting("soundcloud_auth_token", "")
        jamendo_client_id = Settings.get_setting("jamendo_client_id", "")

        subprocess.call(
            [
                "sudo",
                "/usr/local/sbin/raveberry/update_mopidy_config",
                output,
                spotify_username,
                spotify_password,
                spotify_client_id,
                spotify_client_secret,
                soundcloud_auth_token,
                jamendo_client_id,
            ]
        )
        time.sleep(3)

    def __init__(self, settings: Settings):
        self.settings = settings

    def check_mopidy_extensions(self) -> Dict[str, Tuple[bool, str]]:
        """Returns a dict indicating for each extension whether it is enabled
        and provides a message.
        Handles both service and user mopidy instances."""
        try:
            subprocess.check_call(
                ["systemctl", "is-active", "mopidy"], stdout=subprocess.DEVNULL
            )
        except subprocess.CalledProcessError:
            extensions = self._check_mopidy_extensions_user()
        else:
            extensions = self._check_mopidy_extensions_service()
        return extensions

    def _check_mopidy_extensions_user(self) -> Dict[str, Tuple[bool, str]]:
        config = subprocess.run(
            ["mopidy", "config"],
            stdout=subprocess.PIPE,
            universal_newlines=True,
            check=True,
        ).stdout
        parser = configparser.ConfigParser()
        parser.read_string(config)
        extensions = {}
        for extension in ["spotify", "soundcloud", "jamendo"]:
            try:
                if parser[extension]["enabled"] == "true":
                    extensions[extension] = (True, "Extension probably functional")
                else:
                    extensions[extension] = (False, "Extension disabled")
            except KeyError:
                extensions[extension] = (False, "Extension disabled")
        return extensions

    def _check_mopidy_extensions_service(self) -> Dict[str, Tuple[bool, str]]:
        # check the mopidy log and see if there is a spotify login error since the last restart
        log = subprocess.check_output(
            ["sudo", "/usr/local/sbin/raveberry/read_mopidy_log"],
            universal_newlines=True,
        )

        extensions = {}
        for line in log.split("\n")[::-1]:
            if "spotify" not in extensions:

                if (
                    line.startswith("ERROR")
                    and "spotify.session" in line
                    and "USER_NEEDS_PREMIUM"
                ):
                    extensions["spotify"] = (False, "Spotify Premium is required")
                elif line.startswith("ERROR") and "spotify.session" in line:
                    extensions["spotify"] = (False, "User or Password are wrong")
                elif line.startswith("ERROR") and "mopidy_spotify.web" in line:
                    extensions["spotify"] = (
                        False,
                        "Client ID or Client Secret are wrong or expired",
                    )
                elif (
                    line.startswith("WARNING")
                    and "spotify" in line
                    and "The extension has been automatically disabled" in line
                ):
                    extensions["spotify"] = (False, "Configuration Error")

            if "soundcloud" not in extensions:
                if line.startswith("ERROR") and 'Invalid "auth_token"' in line:
                    extensions["soundcloud"] = (False, "auth_token is invalid")
                elif (
                    line.startswith("WARNING")
                    and "soundcloud" in line
                    and "The extension has been automatically disabled" in line
                ):
                    extensions["soundcloud"] = (False, "Configuration Error")

            if "jamendo" not in extensions:
                if line.startswith("ERROR") and 'Invalid "client_id"' in line:
                    extensions["jamendo"] = (False, "client_id is invalid")
                elif (
                    line.startswith("WARNING")
                    and "jamendo" in line
                    and "The extension has been automatically disabled" in line
                ):
                    extensions["jamendo"] = (False, "Configuration Error")

            if (
                "spotify" in extensions
                and "soundcloud" in extensions
                and "jamendo" in extensions
            ):
                break

            if line.startswith("Started Mopidy music server."):
                if "spotify" not in extensions:
                    extensions["spotify"] = (True, "Login successful")
                if "soundcloud" not in extensions:
                    extensions["soundcloud"] = (True, "auth_token valid")
                if "jamendo" not in extensions:
                    extensions["jamendo"] = (True, "client_id could not be checked")
                break
        else:
            # there were too many lines in the log, could not determine whether there was an error
            if "spotify" not in extensions:
                extensions["spotify"] = (True, "No info found, enabling to be safe")
            if "soundcloud" not in extensions:
                extensions["soundcloud"] = (True, "No info found, enabling to be safe")
            if "jamendo" not in extensions:
                extensions["jamendo"] = (True, "No info found, enabling to be safe")
        return extensions

    @Settings.option
    def disable_events(self, _request: WSGIRequest) -> None:
        """Disable websocket support."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_events"])

    @Settings.option
    def enable_events(self, _request: WSGIRequest) -> None:
        """Enable websocket support."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_events"])

    @Settings.option
    def disable_hotspot(self, _request: WSGIRequest) -> None:
        """Disable the wifi created by Raveberry."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_hotspot"])

    @Settings.option
    def enable_hotspot(self, _request: WSGIRequest) -> None:
        """Enable the wifi created by Raveberry."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_hotspot"])

    @Settings.option
    def disable_wifi_protection(self, _request: WSGIRequest) -> None:
        """Disable password protection of the hotspot, making it public."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_wifi_protection"])

    @Settings.option
    def enable_wifi_protection(self, _request: WSGIRequest) -> None:
        """Enable password protection of the hotspot.
        The password was defined during setup."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_wifi_protection"])

    @Settings.option
    def disable_tunneling(self, _request: WSGIRequest) -> None:
        """Disable forwarding of packets to the other network (probably the internet)."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_tunneling"])

    @Settings.option
    def enable_tunneling(self, _request: WSGIRequest) -> None:
        """Enable forwarding of packets to the other network.
        Enables clients connected to the hotspot to browse the internet (if available)."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_tunneling"])
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_tunneling"])

    @Settings.option
    def disable_remote(self, _request: WSGIRequest) -> None:
        """Disables the connection to an external server."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/disable_remote"])

    @Settings.option
    def enable_remote(self, _request: WSGIRequest) -> None:
        """Enables the connection to an external server.
        Allows this instance to be reachable from a public domain."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/enable_remote"])

    @Settings.option
    def reboot_server(self, _request: WSGIRequest) -> None:
        """Restarts the server."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/reboot_server"])

    @Settings.option
    def reboot_system(self, _request: WSGIRequest) -> None:
        """Reboots the system."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/reboot_system"])

    @Settings.option
    def shutdown_system(self, _request: WSGIRequest) -> None:
        """Shuts down the system."""
        subprocess.call(["sudo", "/usr/local/sbin/raveberry/shutdown_system"])

    @classmethod
    @cachetools.func.ttl_cache(ttl=60 * 60 * 24)
    def _fetch_latest_version(cls) -> Optional[str]:
        """Looks up the newest version number from PyPi and returns it."""
        # https://github.com/pypa/pip/issues/9139
        p = subprocess.run(
            "pip3 install --use-deprecated=legacy-resolver raveberry==nonexistingversion".split(),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        if p.returncode == 2:
            # pip does not now the --use-deprecated=legacy-resolver
            # probably an older version that does not need it
            p = subprocess.run(
                "pip3 install raveberry==nonexistingversion".split(),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )

        # parse the newest verson from pip output
        for line in p.stderr.splitlines():
            if "from versions" in line:
                versions = [re.sub(r"[^0-9.]", "", token) for token in line.split()]
                versions = [version for version in versions if version]
                latest_version = versions[-1]
                return latest_version
        else:
            return None

    @Settings.option
    def get_latest_version(self, _request: WSGIRequest) -> HttpResponse:
        """Returns the newest version number of Raveberry from PyPi."""
        latest_version = self._fetch_latest_version()
        if latest_version is None:
            return HttpResponseBadRequest("Could not determine latest version.")
        return HttpResponse(latest_version)

    @Settings.option
    def get_changelog(self, _request: WSGIRequest) -> HttpResponse:
        """Retreives the changelog and returns it."""
        changelog = requests.get(
            "https://raw.githubusercontent.com/raveberry/raveberry/master/CHANGELOG.md"
        ).text
        return HttpResponse(changelog)

    @Settings.option
    def get_upgrade_config(self, _request: WSGIRequest) -> HttpResponse:
        """Returns the config that will be used for the upgrade."""
        with open(os.path.join(settings.BASE_DIR, "config/raveberry.yaml")) as f:
            config = f.read()
        lines = config.splitlines()
        lines = [line for line in lines if not line.startswith("#")]
        return HttpResponse("\n".join(lines))

    @Settings.option
    def upgrade_raveberry(self, _request: WSGIRequest) -> HttpResponse:
        """Performs an upgrade of raveberry."""

        subprocess.call(["sudo", "/usr/local/sbin/raveberry/start_upgrade_service"])

        return HttpResponse("Upgrading... Look for logs in /var/www/")



"""This module handles playback of music."""

from __future__ import annotations

import logging
import os
import random
import time
from contextlib import contextmanager
from threading import Event
from threading import Lock
from threading import Semaphore
from typing import Iterator, Optional, TYPE_CHECKING

import requests
from django.conf import settings
from django.db import transaction
from django.utils import timezone
from mopidyapi.client import MopidyAPI
from mopidyapi.exceptions import MopidyError

import core.models as models
from core.musiq.song_provider import SongProvider
from core.util import background_thread

if TYPE_CHECKING:
    from core.musiq.musiq import Musiq


class Playback:
    """This class handles music playback."""

    # this has to be a class variable
    # so the manager of the queue can access it without a player object
    # a semaphore that indicates whether any _completed_ songs are available in the queue
    # placeholders do not count towards the internal counter
    queue_semaphore: Semaphore = None  # type: ignore

    def __init__(self, musiq: "Musiq") -> None:
        self.musiq = musiq

        self.queue = models.QueuedSong.objects
        self.alarm_playing: Event = Event()
        self.backup_playing: Event = Event()
        self.running = True

        self.player: MopidyAPI = MopidyAPI(host=settings.MOPIDY_HOST)
        self.player_lock = Lock()

    def start(self) -> None:
        self.queue.delete_placeholders()
        Playback.queue_semaphore = Semaphore(self.queue.count())

        with self.mopidy_command(important=True):
            self.player.playback.stop()
            self.player.tracklist.clear()
            # make songs disappear from tracklist after being played
            self.player.tracklist.set_consume(True)
        self._loop()

    def progress(self) -> float:
        """Returns how far into the current song the playback is, in percent."""
        # the state is either pause or stop
        current_position = 0
        duration = 1
        with self.mopidy_command() as allowed:
            if allowed:
                current_position = self.player.playback.get_time_position()
                current_track = self.player.playback.get_current_track()
                if current_track is None or self.backup_playing.is_set():
                    return 0
                duration = current_track.length
        return 100 * current_position / duration

    def paused(self) -> bool:
        """Returns whether playback is currently paused."""
        # the state is either pause or stop
        paused = False
        with self.mopidy_command() as allowed:
            if allowed:
                paused = self.player.playback.get_state() != "playing"
        return paused

    @background_thread
    def _loop(self) -> None:
        """The main loop of the player.
        Takes a song from the queue and plays it until it is finished."""
        while True:

            catch_up = None
            if models.CurrentSong.objects.exists():
                # recover interrupted song from database
                current_song = models.CurrentSong.objects.get()

                # continue with the current song (approximately) where we last left
                song_provider = SongProvider.create(
                    self.musiq, external_url=current_song.external_url
                )
                duration = int(song_provider.get_metadata()["duration"])
                catch_up = round(
                    (timezone.now() - current_song.created).total_seconds() * 1000
                )
                if catch_up > duration * 1000:
                    catch_up = -1
            else:
                self.queue_semaphore.acquire()
                if not self.running:
                    break

                if self.backup_playing.is_set():
                    # stop backup stream
                    self.backup_playing.clear()
                    with self.mopidy_command(important=True) as allowed:
                        if allowed:
                            self.player.playback.next()

                # select the next song depending on settings
                song: Optional[models.QueuedSong]
                if self.musiq.base.settings.basic.voting_system:
                    with transaction.atomic():
                        song = self.queue.confirmed().order_by("-votes", "index")[0]
                        song_id = song.id
                        self.queue.remove(song.id)
                elif self.musiq.controller.shuffle:
                    confirmed = self.queue.confirmed()
                    index = random.randint(0, confirmed.count() - 1)
                    song_id = confirmed[index].id
                    song = self.queue.remove(song_id)
                else:
                    # move the first song in the queue into the current song
                    song_id, song = self.queue.dequeue()

                if song is None:
                    # either the semaphore didn't match up with the actual count
                    # of songs in the queue or a race condition occured
                    logging.warning("dequeued on empty list")
                    continue

                current_song = models.CurrentSong.objects.create(
                    queue_key=song_id,
                    manually_requested=song.manually_requested,
                    votes=song.votes,
                    internal_url=song.internal_url,
                    external_url=song.external_url,
                    stream_url=song.stream_url,
                    artist=song.artist,
                    title=song.title,
                    duration=song.duration,
                )

                self.handle_autoplay()

                try:
                    archived_song = models.ArchivedSong.objects.get(
                        url=current_song.external_url
                    )
                    votes: Optional[int]
                    if self.musiq.base.settings.basic.voting_system:
                        votes = current_song.votes
                    else:
                        votes = None
                    if self.musiq.base.settings.basic.logging_enabled:
                        models.PlayLog.objects.create(
                            song=archived_song,
                            manually_requested=current_song.manually_requested,
                            votes=votes,
                        )
                except (
                    models.ArchivedSong.DoesNotExist,
                    models.ArchivedSong.MultipleObjectsReturned,
                ):
                    pass

            self.musiq.update_state()

            playing = Event()

            @self.player.on_event("playback_state_changed")
            def _on_playback_state_changed(_event):
                playing.set()

            with self.mopidy_command(important=True):
                # after a restart consume may be set to False again, so make sure it is on
                self.player.tracklist.clear()
                self.player.tracklist.set_consume(True)
                self.player.tracklist.add(uris=[current_song.internal_url])
                self.player.playback.play()
                # mopidy can only seek when the song is playing
                started_playing = playing.wait(timeout=1)
                if catch_up is not None and catch_up >= 0:
                    self.player.playback.seek(catch_up)

                # needs some more testing but could prevent "eating the queue" bug
                # if not started_playing and not settings.DOCKER:
                #    logging.error(
                #        "Mopidy did not start playing the song in the timeout. Error assumed.\n"
                #        "Usually faulty outputs are the reason, config is reset to local output.\n"
                #        "Consult mopidy's log for more information.\n"
                #    )
                #    # reset to pulse device 0
                #    self.musiq.base.settings.sound._set_output("0")
                #    # recover the current song by restarting the loop
                #    continue

            self.musiq.update_state()

            if catch_up is None or catch_up >= 0:
                if not self._wait_until_song_end():
                    # there was a ConnectionError during waiting for the song to end
                    # we do not delete the current song but recover its state by restarting the loop
                    continue

            current_song.delete()

            if self.musiq.controller.repeat:
                song_provider = SongProvider.create(
                    self.musiq, external_url=current_song.external_url
                )
                self.queue.enqueue(song_provider.get_metadata(), False)
                self.queue_semaphore.release()

            if (
                self.musiq.base.user_manager.partymode_enabled()
                and random.random() < self.musiq.base.settings.basic.alarm_probability
            ):
                self.alarm_playing.set()
                self.musiq.base.lights.alarm_started()

                self.musiq.update_state()

                with self.mopidy_command(important=True):
                    self.player.tracklist.add(
                        uris=[
                            "file://"
                            + os.path.join(settings.BASE_DIR, "config/sounds/alarm.m4a")
                        ]
                    )
                    self.player.playback.play()
                playing.clear()
                playing.wait(timeout=1)
                self._wait_until_song_end()

                self.musiq.base.lights.alarm_stopped()
                self.musiq.update_state()
                self.alarm_playing.clear()

            if not self.queue.exists() and self.musiq.base.settings.sound.backup_stream:
                self.backup_playing.set()
                # play backup stream
                self.player.tracklist.add(
                    uris=[self.musiq.base.settings.sound.backup_stream]
                )
                self.player.playback.play()

            self.musiq.update_state()

    def _wait_until_song_end(self) -> bool:
        """Wait until the song is over.
        Returns True when finished without errors, False otherwise."""
        # This is the event based approach. Unfortunately too error-prone.
        # playback_ended = Event()
        # @self.player.on_event('tracklist_changed')
        # def on_tracklist_change(event):
        #     playback_ended.set()
        # playback_ended.wait()
        error = False
        while True:
            with self.mopidy_command() as allowed:
                if allowed:
                    try:
                        if self.player.playback.get_state() == "stopped":
                            break
                    except (requests.exceptions.ConnectionError, MopidyError):
                        # error during state get, skip until reconnected
                        error = True
            time.sleep(0.1)
        return not error

    def handle_autoplay(self, url: Optional[str] = None) -> None:
        """Checks whether to add a song by autoplay and does so if necessary.
        :param url: if given, this url is used to find the next autoplayed song.
        Otherwise, the current song is used."""
        if self.musiq.controller.autoplay and models.QueuedSong.objects.count() == 0:
            if url is None:
                # if no url was specified, use the one of the current song
                try:
                    current_song = models.CurrentSong.objects.get()
                    url = current_song.external_url
                except (
                    models.CurrentSong.DoesNotExist,
                    models.CurrentSong.MultipleObjectsReturned,
                ):
                    return

            provider = SongProvider.create(self.musiq, external_url=url)
            try:
                suggestion = provider.get_suggestion()
                # The player loop is not restarted after error automatically.
                # As this function can raise several exceptions (it might do networking)
                # we catch every exception to make sure the loop keeps running
            except Exception as e:  # pylint: disable=broad-except
                logging.exception("error during suggestions for %s: %s", url, e)
            else:
                self.musiq.do_request_music(
                    "",
                    suggestion,
                    None,
                    False,
                    provider.type,
                    archive=False,
                    manually_requested=False,
                )

    @contextmanager
    def mopidy_command(self, important: bool = False) -> Iterator[bool]:
        """A context that should be used around every mopidy command used.
        Makes sure that commands occur sequentially, as mopidy can not handle parallel inputs.
        Use it like this:
        with self.mopidy_command() as allowed:
            if allowed:
                # mopidy command
        :param important: If True, wait until the lock is released.
        If not, return 'False' after a timeout."""
        timeout = 3
        if important:
            timeout = -1
        if self.player_lock.acquire(timeout=timeout):
            yield True
            self.player_lock.release()
        else:
            logging.warning("mopidy command could not be executed")
            yield False

    def start_loop(self) -> None:
        """Starts the playback main loop, only used for tests."""
        if self.running:
            return
        self.running = True
        self._loop()

    def stop_loop(self) -> None:
        """Stops the playback main loop, only used for tests."""
        if not self.running:
            return
        self.running = False
        self.queue_semaphore.release()



"""This module contains all url endpoints and maps them to their corresponding functions."""
import inspect

from typing import Any, List, get_type_hints

from django.core.handlers.wsgi import WSGIRequest
from django.urls import include, URLPattern
from django.urls import path
from django.views.generic import RedirectView

from core import mock
from main import settings

from core.base import Base

BASE = Base()

urlpatterns = [
    path("", RedirectView.as_view(pattern_name="musiq", permanent=False), name="base"),
    path("musiq/", BASE.musiq.index, name="musiq"),
    path("lights/", BASE.lights.index, name="lights"),
    path("stream/", BASE.no_stream, name="no-stream"),
    path("network_info/", BASE.network_info.index, name="network-info"),
    path("settings/", BASE.settings.index, name="settings"),
    path("accounts/", include("django.contrib.auth.urls")),
    path("login/", RedirectView.as_view(pattern_name="login", permanent=False)),
    path("logged_in/", BASE.logged_in, name="logged-in"),
    path("logout/", RedirectView.as_view(pattern_name="logout", permanent=False)),
    path(
        "api/",
        include(
            [
                path(
                    "musiq/",
                    include([path("post_song/", BASE.api.post_song, name="post-song")]),
                )
            ]
        ),
    ),
]

urlmethods = [urlpattern.callback for urlpattern in urlpatterns]


def get_paths(objs: List[Any]) -> List[URLPattern]:
    """Iterates through the given objects and identifies all methods that serve http requests.
    A url pattern is generated for each of these methods,
    but only if no such path exists already in urlpatterns.
    Returns the list of url patterns."""
    paths = []
    for obj in objs:
        for name, method in inspect.getmembers(obj, inspect.ismethod):
            if name == "get_state":
                # the state url is an exception
                # it cannot be named after its method name, as every page has its own get_state
                continue
            if name.startswith("_"):
                # do not expose internal methods
                continue
            type_hints = get_type_hints(method)
            if "request" in type_hints:
                request_type = type_hints["request"]
            elif "_request" in type_hints:
                request_type = type_hints["_request"]
            else:
                continue
            if issubclass(request_type, WSGIRequest) and method not in urlmethods:
                name = name.replace("_", "-")
                paths.append(path(name + "/", method, name=name))
    return paths


base_paths = get_paths([BASE])
BASE.urlpatterns = base_paths
musiq_paths = get_paths([BASE.musiq, BASE.musiq.controller, BASE.musiq.suggestions])
BASE.musiq.urlpatterns = musiq_paths
lights_paths = get_paths([BASE.lights.controller])
BASE.lights.urlpatterns = lights_paths
settings_paths = get_paths(
    [
        BASE.settings.basic,
        BASE.settings.platforms,
        BASE.settings.sound,
        BASE.settings.wifi,
        BASE.settings.library,
        BASE.settings.analysis,
        BASE.settings.system,
    ]
)
BASE.settings.urlpatterns = settings_paths

urlpatterns.append(
    path(
        "ajax/",
        include(
            [
                path("", include(base_paths)),
                path("musiq/state/", BASE.musiq.get_state, name="musiq-state"),
                path("musiq/", include(musiq_paths)),
                path("lights/state/", BASE.lights.get_state, name="lights-state"),
                path("lights/", include(lights_paths)),
                path("settings/state/", BASE.settings.get_state, name="settings-state"),
                path("settings/", include(settings_paths)),
            ]
        ),
    )
)

if settings.MOCK:
    for url in urlpatterns:
        if isinstance(url, URLPattern):
            url.callback = mock.index
else:
    BASE.start()
